import java.util.Map;
import java.util.HashMap;
import java.util.Arrays;
import java.io.*;

public class H89FloppyFormatter implements DiskImage {

	public static final int MMS  = 1;
	public static final int Z17  = 2;
	public static final int M47  = 3;
	public static final int Z37  = 4;
	public static final int Z47  = 5;
	public static final int Z67  = 6;
	public static final int Z37X  =7;
	public static final int Z47X  =8;
	public static final int KAYPRO=9;
	public static final int HIDENS=10;
	public static final int FMT_MASK = 0x00f;

	public static final int F525 = 0x100;
	public static final int F8   = 0x200;
	public static final int MED_MASK = 0xf00;

	public static final int SD = 0x10;
	public static final int DD = 0x20;
	public static final int QD = 0x30;
	public static final int DEN_MASK = 0x0f0;

	public static String[] fmts = new String[] {
		/* none */ "NONE",
		/* MMS */ "MMS",
		/* Z17 */ "Z17",
		/* M47 */ "M47",
		/* Z37 */ "Z37",
		/* Z47 */ "Z47",
		/* Z67 */ "Z67",
		/* Z37X */ "Z37X",
		/* Z47X */ "Z47X",
		/* Kaypro */ "KAYPRO",
		/* Kaypro QT */ "HIDENS",
	};

	// Index bits are: sd, st, ss...
	private static String[] dtddds = new String[] {
		/* stsdss */ "ST SD SS",
		/* stsdds */ "ST SD DS",
		/* dtsdss */ "DT SD SS",
		/* dtsdds */ "DT SD DS",
		/* stddss */ "ST DD SS",
		/* stddds */ "ST DD DS",
		/* dtddss */ "DT DD SS",
		/* dtddds */ "DT DD DS",
		/* qtqdds */ "QT QD DS",
		/* qtqdds */ "QT QD DS",
		/* qtqdds */ "QT QD DS",
		/* qtqdds */ "QT QD DS",
		/* qtqdds */ "QT QD DS",
		/* qtqdds */ "QT QD DS",
		/* qtqdds */ "QT QD DS",
		/* qtqdds */ "QT QD DS",
	};

	static int[][] knownDpbs = new int[][] {
/*  0: 8,M47,DD,SS,ST */	{ 64, 4, 15, 0, 299, 191, 0xe000, 48, 2 },
/*  1: 8,M47,DD,DS,ST */	{ 64, 4, 15, 0, 607, 191, 0xe000, 48, 2 },

/*  2: 5,MMS,DD,SS,ST */	{ 36, 4, 15, 1,  82,  95, 0xc000, 24, 3 },
/*  3: 5,MMS,DD,DS,ST */	{ 36, 4, 15, 1, 172,  95, 0xc000, 24, 3 },
/*  4: 5,MMS,DD,SS,DT */	{ 36, 5, 31, 3,  85, 127, 0x8000, 32, 3 },
/*  5: 5,MMS,DD,DS,DT */	{ 36, 5, 31, 3, 175, 127, 0x8000, 32, 3 },

/*  6: 8,MMS,DD,SS,ST */	{ 64, 4, 15, 0, 299, 191, 0xe000, 48, 2 },
/*  7: 8,MMS,DD,DS,ST */	{ 64, 4, 15, 0, 607, 191, 0xe000, 48, 2 },

/*  8: 5,Z37,SD,SS,ST */	{ 20, 3,  7, 0,  91,  63, 0xc000, 16, 3 },
/*  9: 5,Z37,SD,DS,ST */	{ 20, 3,  7, 0, 187, 127, 0xf000, 32, 3 },
/* 10: 5,Z37,SD,SS,DT */	{ 20, 3,  7, 0, 191,  63, 0xc000, 16, 3 },
/* 11: 5,Z37,SD,DS,DT */	{ 20, 4, 15, 1, 195, 127, 0xc000, 32, 3 },

/* 12: 5,Z37,DD,SS,ST */	{ 32, 4, 15, 1, 151, 127, 0xf000, 32, 2 },
/* 13: 5,Z37,DD,DS,ST */	{ 32, 4, 15, 0, 155, 255, 0xf000, 64, 2 },
/* 14: 5,Z37,DD,SS,DT */	{ 32, 4, 15, 1, 155, 127, 0xc000, 32, 2 },
/* 15: 5,Z37,DD,DS,DT */	{ 32, 4, 15, 0, 315, 255, 0xf000, 64, 2 },

/* 16: 5,Z37X,DD,SS,ST */	{ 40, 3,  7, 0, 185, 127, 0xf000, 32, 2 },
/* 17: 5,Z37X,DD,DS,ST */	{ 40, 4, 15, 0, 194, 255, 0xf000, 64, 2 },
/* 18: 5,Z37X,DD,SS,DT */	{ 40, 4, 15, 1, 194, 127, 0xc000, 32, 2 },
/* 19: 5,Z37X,DD,DS,DT */	{ 40, 4, 15, 0, 394, 255, 0xf000, 64, 2 },

/* 20: 8,Z47,SD,SS,ST */	{ 26, 3,  7, 0, 242,  63, 0xc000, 16, 2 },
/* 21: 8,Z47,SD,DS,ST */	{ 26, 4, 15, 1, 246, 127, 0xc000, 32, 2 },
/* 22: 8,Z47,DD,SS,ST */	{ 52, 4, 15, 0, 242, 127, 0xc000, 32, 2 },
/* 23: 8,Z47,DD,DS,ST */	{ 52, 4, 15, 0, 493, 255, 0xf000, 64, 2 },

/* 24: 8,Z47X,DD,SS,ST */	{ 64, 4, 15, 0, 299, 127, 0xc000, 32, 2 },
/* 25: 8,Z47X,DD,DS,ST */	{ 64, 4, 15, 0, 607, 255, 0xf000, 64, 2 },

/* 26: 5,Z17,SD,SS,ST */	{ 20, 3,  7, 0,  91,  63, 0xc000, 16, 3 },
/* 27: 5,Z17,SD,DS,ST */	{ 20, 3,  7, 0, 181,  63, 0xc000, 16, 3 },
/* 28: 5,Z17,SD,SS,DT */	{ 20, 4, 15, 1,  95, 127, 0xc000, 32, 3 },
/* 29: 5,Z17,SD,DS,DT */	{ 20, 4, 15, 1, 185, 127, 0xc000, 32, 3 },

/* 30: 5,KAYPRO,DD,SS,ST */	{ 40, 3,  7, 0, 194,  63, 0xf000, 16, 1 },
/* 31: 5,KAYPRO,DD,DS,ST */	{ 40, 4, 15, 1, 196,  63, 0xc000, 16, 1 },

/* 32: 5,HIDENS,QD,DS,QT */	{ 68, 4, 15, 0,1351,1023, 0xffff,256, 2 },

	};

	static Map<Integer, Integer[]> geometries = new HashMap<Integer, Integer[]>();
	static {
		// latency is only relavent for hard-sectored disks.
		// soft sectored disk images directly access sector requested,
		// no translation needed.
		// negative latency indicates physical formatted sector skew,
		// OS does no translation.
		//                                             --physical---
		//                                             spt   ssz lat dpbx,itl
		geometries.put(F525 | Z17 | SD,  new Integer[]{ 10,  256,  4,26,0 });
		geometries.put(F525 | MMS | DD,  new Integer[]{  9,  512, -7, 2,0 });
		geometries.put(F525 | Z37 | SD,  new Integer[]{ 10,  256, -7, 8,1 });
		geometries.put(F525 | Z37 | DD,  new Integer[]{ 16,  256,-11,12,1 });
		geometries.put(F525 | Z37X | DD, new Integer[]{  5, 1024, -2,16,1 });
		geometries.put(F8 | MMS | DD,    new Integer[]{ 16,  512,-13, 6,0 });
		geometries.put(F8 | M47 | DD,    new Integer[]{  8, 1024,  4, 0,1 });
		geometries.put(F8 | Z47 | DD,    new Integer[]{ 26,  256,  9,22,1 });
		geometries.put(F8 | Z47 | SD,    new Integer[]{ 26,  128,  6,20,1 });
		geometries.put(F8 | Z47X | DD,   new Integer[]{  8, 1024,  4,24,1 });
		geometries.put(F8 | Z67 | DD,    new Integer[]{ 26,  256,  9,22,1 });
		geometries.put(F8 | Z67 | SD,    new Integer[]{ 26,  128,  6,20,1 });
		geometries.put(F525 | KAYPRO | DD,new Integer[]{10,  512, -1,30,2 });
		geometries.put(F525 | HIDENS | QD,new Integer[]{17,  512,  5,32,1 });
	}

	int spt;
	int hspt;
	int ssz;
	int lat;
	int fmt;
	int med;
	int den;
	int ntrk;
	int nsid;
	int tlen;
	int itl;
	long cap;
	public CpmDpb dpb = null;

	private static void printOne(int fmt, int med, int ssz, int den, int dsdt, int[] dpb) {
		if (den == DD) {
			dsdt |= 4;
		}
		if (den == QD) {
			dsdt |= 8;
		}
		System.out.format("%4s %s %4d %2d %d %2d %d %3d %3d %s %d %d %d\n",
			fmts[fmt], dtddds[dsdt], ssz, dpb[0], dpb[1], dpb[2], dpb[3],
			dpb[4], dpb[5], Integer.toString(dpb[6], 2),
			dpb[7], dpb[8], med == F525 ? 5 : 8);
	}

	public static void list() {
		System.out.format(" fmt tr de si  ssz -------------- CP/M DPB --------------- media\n");
		for (int fmt : geometries.keySet()) {
			Integer[] geom = geometries.get(fmt);
			int dpbx = geom[3];
			int f = fmt & FMT_MASK;
			int m = fmt & MED_MASK;
			int d = fmt & DEN_MASK;
			int x = 0;
			if (m == F525) {
				x = 4;
			} else {
				x = 2;
			}
			for (int y = 0; y < x; ++y) {
				printOne(f, m, (int)geom[1], d, y, knownDpbs[dpbx + y]);
			}
		}
	}

	public H89FloppyFormatter(int format, boolean ds, boolean dt) {
		fmt = format & FMT_MASK;
		if (fmt == HIDENS) {
			format = HIDENS | F525 | QD;
		}
		Integer[] geom = geometries.get(format);
		if (geom == null) {
			spt = 0;
			return;
		}
		med = format & MED_MASK;
		den = format & DEN_MASK;
		spt = geom[0];
		ssz = geom[1];
		lat = geom[2];
		if (lat < 1) {
			// We don't physically format sector ordering...
			lat = 1;
		}
		itl = geom[4];
		int dpbx = geom[3];
		if (dpbx >= 0) {
			if (fmt != HIDENS) {
				if (ds) {
					dpbx += 1;
				}
				if (dt) {
					dpbx += 2;
				}
			}
			dpb = new CpmDpb(knownDpbs[dpbx]);
		}
		if (fmt == Z17) { // TODO: abstract hard-sectored?
			hspt = spt; // num holes excl index
		} else {
			hspt = 0; // only index hole
		}
		if (lat > 1) {
			dpb.pssz = ssz;
			dpb.plat = lat;
		}
		if (fmt == HIDENS) {
			tlen = 9600; // default DD.
			ntrk = 160;
		} else if (med == F525) {
			tlen = 6400; // default DD.
			ntrk = dt ? 80 : 40;
		} else {
			tlen = 12800; // default DD.
			ntrk = 77;
		}
		if (den != DD) {
			tlen /= 2;
		}
		nsid = ds ? 2 : 1;
		cap = ntrk * nsid * spt * ssz;
		//System.err.format("FMT %d\" %d %d %d %d DD %s INTERLACE %s %d\n",
		//med == F525 ? 5 : 8, nsid, ntrk, spt, ssz, (den == DD), (itl != 0), cap);
	}

	public boolean isError() { return spt == 0; }

	public long getCapacity() { return cap; }

	public boolean format(File fi) {
		try {
			int trk, sid, sec;
			byte[] buf = new byte[ssz];
			Arrays.fill(buf, (byte)0xe5);
			OutputStream fd = new FileOutputStream(fi);
			// For a blank disk, interlace doesn't matter.
			for (trk = 0; trk < ntrk; ++trk) {
				for (sid = 0; sid < nsid; ++sid) {
					for (sec = 0; sec < spt; ++sec) {
						fd.write(buf);
					}
				}
			}
			Arrays.fill(buf, (byte)0);
			String hdr = String.format("%dm%dz%dp%ds%dt%dd%di%dl%dh\n",
				med == F525 ? 5 : 8, ssz, spt, nsid, ntrk,
				den == QD ? 2 : den == DD ? 1 : 0, itl, lat, hspt);
			byte[] b = hdr.getBytes();
			System.arraycopy(b, 0, buf, 0, b.length);
			fd.write(buf, 0, 128);
			fd.close();
			return true;
		} catch (Exception ee) {
			ee.printStackTrace();
			return false;
		}
	}

	public boolean isCompatible(DiskImage di) {
		int m = med == F525 ? 5 : 8;
		// TODO: add ntrk and nsid...
		return m == di.mediaSize() &&
			spt == di.sectorsPerTrack() &&
			ssz == di.sectorSize() &&
			(itl != 0) == di.interlaced() &&
			cap == di.capacity();
	}

	public int mediaSize() { return med == F525 ? 5 : 8; }
	public int numSides() { return nsid; }
	public int numTracks() { return ntrk; }
	public int sectorsPerTrack() { return spt; }
	public int sectorSize() { return ssz; }
	public boolean interlaced() { return itl != 0; }
	public int capacity() { return (int)cap; }
}
