import java.util.Arrays;
import java.io.*;

public class H89SASIFormatter {
	// Taken from GenericSASIDrive. Keep in sync!
	public static final int XEBEC_ST506 = 1;	//  5.3MB (XEBEC ctrlr + Seagate ST506 drive)
	public static final int XEBEC_ST412 = 2;	// 10.6MB
	public static final int XEBEC_CM5206 = 3;	//  4.4MB
	public static final int XEBEC_CM5410 = 4;	//  8.9MB
	public static final int XEBEC_CM5616 = 5;	// 13.3MB
	public static final int XEBEC_RO201 = 6;	//  5.5MB
	public static final int XEBEC_RO202 = 7;	// 11.1MB
	public static final int XEBEC_RO203 = 8;	// 16.7MB
	public static final int XEBEC_RO204 = 9;	// 22.3MB
	private int[][] params = new int[][] {
		{ 0, 0, 0, 0 },		// /* INVALID=0    */
		{ 153, 4, 128, 64 },	// /*[XEBEC_ST506] */
		{ 306, 4, 128, 64 },	// /*[XEBEC_ST412] */
		{ 256, 2, 256, 256 },	// /*[XEBEC_CM5206]*/
		{ 256, 4, 256, 256 },	// /*[XEBEC_CM5410]*/
		{ 256, 6, 256, 256 },	// /*[XEBEC_CM5616]*/
		{ 321, 2, 132, 0 },	// /*[XEBEC_RO201] */
		{ 321, 4, 132, 0 },	// /*[XEBEC_RO202] */
		{ 321, 6, 132, 0 },	// /*[XEBEC_RO203] */
		{ 321, 8, 132, 0 },	// /*[XEBEC_RO204] */
	};
	private boolean checkHeader(byte[] buf) {
		int m = 0;
		int e = 0;
		while (buf[e] != '\n' && buf[e] != '\0' && e < buf.length) {
			int p = 0;
			while (e < buf.length && Character.isDigit(buf[e])) {
				p = (p * 10) + (buf[e] - '0');
				++e;
			}
			if (e >= buf.length) {
				// too harsh?
				return false;
			}
			// TODO: removable flag, others?
			// NOTE: removable media requires many more changes.
			switch (Character.toLowerCase(buf[e])) {
			case 'c':
				if (p > 0x0ffff) {
					break;
				}
				m |= 0x01;
				cyls = p & 0x0ffff;
				break;

			case 'h':
				if (p > 0x0f) {
					break;
				}
				m |= 0x02;
				hds = p & 0x0f;
				break;

			case 'z':
				if (p != 128 && p != 256 && p !=512 && p != 1024) {
					break;
				}
				m |= 0x04;
				ssz = p & 0x0fff;
				break;

			case 'p':
				if (p > 0x0ff) {
					break;
				}
				m |= 0x08;
				spt = p & 0x0ff;
				break;

			case 'l':
				if (p > 0x0ff) {
					break;
				}
				m |= 0x10;
				lat = p & 0x0ff;
				break;

			default:
				return false;
			}

			e++;
		}
		return m == 0x1f;
	}
	// END GenericSASIDrive code
	public static String[] disks;

	public static final int RESV_SECS = 12;

	static private byte[] boot_jmp = new byte[] {
		(byte)0xC3, (byte)0x80, (byte)0x24
	};
	static private byte[] boot = new byte[] {
(byte)0xC3, (byte)0x85, (byte)0x24, (byte)0x00, (byte)0x00, (byte)0x3B, (byte)0x3B, (byte)0xD1,
(byte)0x21, (byte)0xF0, (byte)0x25, (byte)0xC1, (byte)0x79, (byte)0xFE, (byte)0xC3, (byte)0x28,
(byte)0x03, (byte)0x77, (byte)0x78, (byte)0xB7, (byte)0x31, (byte)0x80, (byte)0x26, (byte)0xD5,
(byte)0xC0, (byte)0x3A, (byte)0x34, (byte)0x20, (byte)0xFE, (byte)0x05, (byte)0x30, (byte)0x04,
(byte)0xD6, (byte)0x03, (byte)0x18, (byte)0x02, (byte)0xD6, (byte)0xA8, (byte)0xE6, (byte)0x03,
(byte)0x0F, (byte)0x0F, (byte)0x0F, (byte)0x32, (byte)0xEA, (byte)0x25, (byte)0x3A, (byte)0x8D,
(byte)0x22, (byte)0x32, (byte)0xE9, (byte)0x25, (byte)0x3A, (byte)0x91, (byte)0x22, (byte)0x32,
(byte)0xED, (byte)0x25, (byte)0x3A, (byte)0x50, (byte)0x21, (byte)0x4F, (byte)0x0C, (byte)0xCD,
(byte)0x6C, (byte)0x25, (byte)0xCD, (byte)0x8D, (byte)0x25, (byte)0xCD, (byte)0xCB, (byte)0x25,
(byte)0x3A, (byte)0x83, (byte)0x22, (byte)0xE6, (byte)0xE0, (byte)0xC2, (byte)0xE9, (byte)0x24,
(byte)0x0C, (byte)0x3E, (byte)0x0C, (byte)0x32, (byte)0xE9, (byte)0x25, (byte)0xCD, (byte)0x6C,
(byte)0x25, (byte)0xCD, (byte)0x8D, (byte)0x25, (byte)0x21, (byte)0x85, (byte)0x22, (byte)0x06,
(byte)0x08, (byte)0x1E, (byte)0xC8, (byte)0xCD, (byte)0x95, (byte)0x25, (byte)0xCD, (byte)0xCB,
(byte)0x25, (byte)0x21, (byte)0xF0, (byte)0x25, (byte)0x7E, (byte)0xB7, (byte)0x28, (byte)0x07,
(byte)0xD6, (byte)0x30, (byte)0x21, (byte)0x93, (byte)0x22, (byte)0xBE, (byte)0xD0, (byte)0x21,
(byte)0x94, (byte)0x22, (byte)0x4F, (byte)0x06, (byte)0x00, (byte)0x09, (byte)0x09, (byte)0x09,
(byte)0xE5, (byte)0x21, (byte)0x9A, (byte)0x22, (byte)0x11, (byte)0x15, (byte)0x00, (byte)0x19,
(byte)0x0D, (byte)0xF2, (byte)0x07, (byte)0x25, (byte)0x11, (byte)0x0F, (byte)0x00, (byte)0x19,
(byte)0x7E, (byte)0xE1, (byte)0x4E, (byte)0x23, (byte)0x5E, (byte)0x23, (byte)0x56, (byte)0xE6,
(byte)0x03, (byte)0x47, (byte)0x32, (byte)0x83, (byte)0x24, (byte)0x28, (byte)0x13, (byte)0x79,
(byte)0xE6, (byte)0x1F, (byte)0x4F, (byte)0x3E, (byte)0x01, (byte)0xCB, (byte)0x19, (byte)0xCB,
(byte)0x1B, (byte)0xCB, (byte)0x1A, (byte)0x07, (byte)0xB7, (byte)0x10, (byte)0xF6, (byte)0x32,
(byte)0x84, (byte)0x24, (byte)0x3A, (byte)0xEA, (byte)0x25, (byte)0xB1, (byte)0x32, (byte)0xEA,
(byte)0x25, (byte)0xED, (byte)0x53, (byte)0xEB, (byte)0x25, (byte)0x3E, (byte)0x01, (byte)0x32,
(byte)0xED, (byte)0x25, (byte)0x3A, (byte)0x84, (byte)0x22, (byte)0x32, (byte)0xEE, (byte)0x25,
(byte)0x3E, (byte)0x08, (byte)0x32, (byte)0xE9, (byte)0x25, (byte)0x3A, (byte)0x50, (byte)0x21,
(byte)0x4F, (byte)0x0C, (byte)0xCD, (byte)0x6C, (byte)0x25, (byte)0xCD, (byte)0x8D, (byte)0x25,
(byte)0xCD, (byte)0xAE, (byte)0x25, (byte)0xCD, (byte)0xCB, (byte)0x25, (byte)0x21, (byte)0xEA,
(byte)0x25, (byte)0x11, (byte)0x77, (byte)0x23, (byte)0x01, (byte)0x05, (byte)0x00, (byte)0xED,
(byte)0xB0, (byte)0xC3, (byte)0x80, (byte)0x22, (byte)0x06, (byte)0x00, (byte)0xED, (byte)0x78,
(byte)0xE6, (byte)0x08, (byte)0x28, (byte)0x04, (byte)0x10, (byte)0xF8, (byte)0x18, (byte)0x34,
(byte)0x3E, (byte)0x40, (byte)0xED, (byte)0x79, (byte)0x06, (byte)0x00, (byte)0xED, (byte)0x78,
(byte)0xE6, (byte)0x08, (byte)0x20, (byte)0x04, (byte)0x10, (byte)0xF8, (byte)0x18, (byte)0x24,
(byte)0x3E, (byte)0x00, (byte)0xED, (byte)0x79, (byte)0xC9, (byte)0x0D, (byte)0x21, (byte)0xE9,
/* end 256 bytes */
(byte)0x25, (byte)0x06, (byte)0x06, (byte)0x1E, (byte)0xD8, (byte)0xC5, (byte)0x0C, (byte)0x06,
(byte)0x10, (byte)0xED, (byte)0x78, (byte)0xE6, (byte)0xD8, (byte)0xBB, (byte)0x28, (byte)0x05,
(byte)0x10, (byte)0xF7, (byte)0xC1, (byte)0x18, (byte)0x07, (byte)0xC1, (byte)0xED, (byte)0xA3,
(byte)0xC2, (byte)0x95, (byte)0x25, (byte)0xC9, (byte)0xD1, (byte)0xC9, (byte)0x21, (byte)0x80,
(byte)0x22, (byte)0x0C, (byte)0xED, (byte)0x78, (byte)0x0D, (byte)0xE6, (byte)0xD8, (byte)0xFE,
(byte)0x98, (byte)0xC8, (byte)0xE6, (byte)0x98, (byte)0xFE, (byte)0x88, (byte)0x20, (byte)0xF1,
(byte)0x3A, (byte)0x84, (byte)0x24, (byte)0x06, (byte)0x80, (byte)0xED, (byte)0xB2, (byte)0x3D,
(byte)0x20, (byte)0xF9, (byte)0xC9, (byte)0x21, (byte)0xEF, (byte)0x25, (byte)0x18, (byte)0x03,
(byte)0xED, (byte)0x78, (byte)0x77, (byte)0x0C, (byte)0xED, (byte)0x78, (byte)0x0D, (byte)0xE6,
(byte)0xF0, (byte)0xFE, (byte)0x90, (byte)0x28, (byte)0xF3, (byte)0xFE, (byte)0xB0, (byte)0x20,
(byte)0xF2, (byte)0xED, (byte)0x78, (byte)0x7E, (byte)0xE6, (byte)0x03, (byte)0x20, (byte)0xC4,
(byte)0xC9, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,
	};

	static private byte[] idat = new byte[RESV_SECS * 128];

	static {
		Arrays.fill(idat, (byte)0);
		System.arraycopy(boot_jmp, 0, idat, 0, boot_jmp.length);
		System.arraycopy(boot, 0, idat, 512, boot.length);
		disks = new String[10];
		disks[XEBEC_ST506] = "XEBEC_ST506";
		disks[XEBEC_ST412] = "XEBEC_ST412";
		disks[XEBEC_CM5206] = "XEBEC_CM5206";
		disks[XEBEC_CM5410] = "XEBEC_CM5410";
		disks[XEBEC_CM5616] = "XEBEC_CM5616";
		disks[XEBEC_RO201] = "XEBEC_RO201";
		disks[XEBEC_RO202] = "XEBEC_RO202";
		disks[XEBEC_RO203] = "XEBEC_RO203";
		disks[XEBEC_RO204] = "XEBEC_RO204";
	}

	int disk = 0;;
	int minBsh; // hint... minimum
	int spt;
	int ssz;
	int cyls;
	int hds;
	public int resv_trks;
	int resv_secs; // 128-byte sectors
	int lat;
	int capacity; // 'ssz'-byte sectors
	float factor; // padding for directory
	int[] parts = null;
	int nparts;
	byte[] sec0;
	CpmDpb[] dpbs = null;
	// These are transient... setcpmdpb() sets.
	int blocksize;
	int maxdir;

	public H89SASIFormatter(File fi) {
		sec0 = new byte[idat.length];
		byte[] buf = new byte[128];
		try {
			RandomAccessFile fd = new RandomAccessFile(fi, "r");
			fd.read(sec0);
			fd.seek(fd.length() - buf.length);
			fd.read(buf);
			fd.close();
		} catch (Exception ee) {
			return;
		}
		if (!checkHeader(buf)) {
			return;
		}
		try {
			for (int dsk = 1; true; ++dsk) {
				// hopefully these are unique...
				if (params[dsk][0] == cyls && params[dsk][1] == hds) {
					disk = dsk;
					break;
				}
			}
		} catch (Exception ee) {}
		if (disk == 0) {
			return;
		}
		capacity = cyls * hds * spt;
		resv_secs = RESV_SECS; // per disk, for init data. 128-byte sectors!
		resv_trks = 2; // each partition, for CP/M...
		parts = new int[8]; // allocate all 8 anyway
		Arrays.fill(parts, 0);
		dpbs = new CpmDpb[8];
		int adr = 19;
		int adr2 = 47;
		nparts = sec0[adr++] & 0xff;
		int t = resv_secs;
		int sxf = ssz / 128;
		for (int x = 0; x < nparts; ++x) {
			// this is the partition start address...
			parts[x] = getpart(sec0, adr);
			if (x > 0) {
				parts[x - 1] = parts[x] - parts[x - 1];
				t += parts[x - 1] * sxf;
			}
			dpbs[x] = new CpmDpb();
			getcpmdpb(sec0, adr2, dpbs[x]);
			dpbs[x].poff = t;
			adr += 3;;
			adr2 += 21;;
		}
		parts[nparts - 1] = capacity * sxf - parts[nparts - 1];
	}

	public H89SASIFormatter(int dsk, int sz, int lt, int mBsh, float fact) {
		if (sz == 512) {
			spt = 17;
		} else if (sz == 256) {
			spt = 32;
		} else {
			return;
		}
		try {
			cyls = params[dsk][0];
			hds = params[dsk][1];
		} catch (Exception ee) {
			// likely "array index out of bounds"...
			return;
		}
		ssz = sz;
		disk = dsk;
		capacity = cyls * hds * spt;
		resv_secs = RESV_SECS; // per disk, for init data. 128-byte sectors!
		resv_trks = 2; // each partition, for CP/M...
		lat = lt; // not used?
		minBsh = mBsh; // minimum value
		factor = fact;
	}

	public void dump(PrintStream out) {
		double cap = getCapacity();
		cap /= 1024;
		cap /= 1024;
		out.format("Drive %s (%.1f MB)\n", disks[disk], cap);
		out.format("Sector Size %d, Sec/Trk %d, Cylinders %d, Heads %d\n",
			ssz, spt, cyls, hds);
		for (int x = 0; x < nparts; ++x) {
			cap = parts[x] * 128;
			cap /= 1024;
			cap /= 1024;
			out.format("Partition %d %dx%d (%.1fMB) CP/M DPB:\n", x,
				parts[x], ssz, cap);
			out.format("    %d spt\n", dpbs[x].spt);
			out.format("    %d,%d,%d bsh,blm,exm\n",
				dpbs[x].bsh, dpbs[x].blm, dpbs[x].exm);
			out.format("    %d,%d dsm,drm\n",
				dpbs[x].dsm, dpbs[x].drm);
			out.format("    %s alv0\n", Integer.toBinaryString(dpbs[x].al0));
			out.format("    %d,%d cks,off\n",
				dpbs[x].cks, dpbs[x].off);
			out.format("    %d [poff]\n", dpbs[x].poff);
		}
	}

	// Total capacity in bytes
	public int getCapacity() { return capacity * ssz; }

	// Available number of sectors (ssz).
	public int getAvailable() { return capacity - (resv_secs * 128) / ssz; }

	// Partition Table (pt[]) is in 'sz' size sectors (not 128-byte).
	// ddfi may be null.
	public boolean setPartitions(int[] pt, int np, File ddfi, CpmDpb[] dpbs) {
		if (isError() || np <= 0 || np > 8) {
			return false;
		}
		String basename = "";
		PrintStream ddfp = null;
		if (ddfi != null) {
			basename = ddfi.getName();
			try {
				ddfp = new PrintStream(ddfi);
			} catch (Exception ee) {}
		}
		sec0 = Arrays.copyOf(idat, idat.length);
		int adr = 3;
		Arrays.fill(sec0, adr, 256, (byte)0);
		sec0[adr++] = (byte)0b00011111;	// drive/ctrlr code TODO: real codes...
		sec0[adr++] = (byte)1;		// control byte
		setdchars(sec0, adr, disk); // 8 bytes...
		adr += 8;
		// assign drive type - not used. 6 bytes.
		adr += 6;
		sec0[adr++] = (byte)np;
		int t = resv_secs;
		int sxf = ssz / 128;
		for (int x = 0; x < 9; ++x) {
			if (x < np) {
				System.out.format("Partition %d start %d\n", x, t);
				setpart(sec0, adr, t);
				t += pt[x] * sxf;
			} else {
				setpart(sec0, adr, 0);
			}
			adr += 3;
		}
		// +47 bytes. Now CP/M DPBs for each partition...  21 bytes each
		t = resv_secs;
		for (int x = 0; x < np; ++x) {
			setcpmdpb(sec0, adr, pt[x], dpbs[x]);
			adr += 21;
			if (ddfp != null) {
				int lspt = 8192 / ssz;
				ddfp.format("diskdef %s.%d.%d\n" +
					"  seclen %d\n  tracks %d\n  sectrk %d\n" +
					"  blocksize %d\n  maxdir %d\n  skew 0\n" +
					"  boottrk 2\n  offset %d\n  os 2.2\nend\n",
					basename, np, x,
					ssz, pt[x] / lspt, lspt,
					blocksize, maxdir,
					t * 128);
			}
			t += pt[x] * sxf;
		}
		return true;
	}

	public boolean isError() { return disk == 0; }

	private void setdb(byte[] buf, int start, int db) {
		buf[start] = (byte)(db & 0x0ff);
	}

	private void setdw(byte[] buf, int start, int dw) {
		// little-endian
		buf[start++] = (byte)(dw & 0x0ff);
		buf[start++] = (byte)((dw >> 8) & 0x0ff);
	}

	private void setd2(byte[] buf, int start, int dw) {
		// big-endian
		buf[start++] = (byte)((dw >> 8) & 0x0ff);
		buf[start++] = (byte)(dw & 0x0ff);
	}

	private void setdchars(byte[] buf, int start, int disk) {
		setd2(buf, start + 0, params[disk][0]);
		setdb(buf, start + 2, params[disk][1]);
		setd2(buf, start + 3, params[disk][2]);
		setd2(buf, start + 5, params[disk][3]);
		buf[start + 7] = (byte)0; // ECC burst length - not relavent
	}

	private int getpart(byte[] buf, int start) {
		// big-endian
		int p = (buf[start++] & 0xff) << 16;
		p |= (buf[start++] & 0xff) << 8;
		p |= buf[start++] & 0xff;
		return p;
	}

	private void setpart(byte[] buf, int start, int pstart) {
		// big-endian
		buf[start++] = (byte)((pstart >> 16) & 0x0ff);
		buf[start++] = (byte)((pstart >> 8) & 0x0ff);
		buf[start++] = (byte)(pstart & 0x0ff);
	}

	private void getcpmdpb(byte[] buf, int start, CpmDpb dpb) {
		dpb.spt = buf[start++] & 0xff;
		dpb.spt |= (buf[start++] & 0xff) << 8;
		dpb.bsh = buf[start++] & 0xff;
		dpb.blm = buf[start++] & 0xff;
		dpb.exm = buf[start++] & 0xff;
		dpb.dsm = buf[start++] & 0xff;
		dpb.dsm |= (buf[start++] & 0xff) << 8;
		dpb.drm = buf[start++] & 0xff;
		dpb.drm |= (buf[start++] & 0xff) << 8;
		dpb.al0 = (buf[start++] & 0xff) << 8; // big-endian
		dpb.al0 |= buf[start++] & 0xff;
		dpb.cks = buf[start++] & 0xff;
		dpb.cks |= (buf[start++] & 0xff) << 8;
		dpb.off = buf[start++] & 0xff;
		dpb.off |= (buf[start++] & 0xff) << 8;
		// caller must set poff...
	}

	private int setcpmdpb(byte[] buf, int start, int psize, CpmDpb dpb) {
		blocksize = (1 << dpb.bsh) * 128;
		maxdir = dpb.drm + 1;
		setdw(buf, start + 0, 64); // SPT - constant, regardless of hardware.
		setdb(buf, start + 2, dpb.bsh);
		setdb(buf, start + 3, dpb.blm);
		setdb(buf, start + 4, dpb.exm);
		setdw(buf, start + 5, dpb.dsm);
		setdw(buf, start + 7, dpb.drm);
		setd2(buf, start + 9, dpb.al0);	// essentially big-endian...
		setdw(buf, start + 11, 0);	// fixed media, no CKS
		setdw(buf, start + 13, dpb.off);
		// non-CP/M:
		setdb(buf, start + 15, 0b00000010);
		setdb(buf, start + 16, 0b10000000);
		setdb(buf, start + 17, 0b00000000);
		setdb(buf, start + 18, 0b11111111);
		setdb(buf, start + 19, 0b11111111);
		setdb(buf, start + 20, 0b11111111);
		return 0;
	}

	public boolean format(File fi, boolean repair) {
		// TODO: handle truncate vs. overwrite... need RandomAccessFile?
		if (sec0 == null) {
			sec0 = Arrays.copyOf(idat, idat.length);
			Arrays.fill(sec0, (byte)0xe5);
		}
		try {
			OutputStream fd = new FileOutputStream(fi);
			fd.write(sec0);

			if (!repair) {
				byte[] buf = new byte[ssz];
				Arrays.fill(buf, (byte)0xe5);
				for (int x = sec0.length / ssz; x < capacity; ++x) {
					fd.write(buf);
				}
				Arrays.fill(buf, (byte)0);
				String hdr = String.format("%dc%dh%dz%dp%dl\n",
					cyls, hds, ssz, spt, lat);
				byte[] b = hdr.getBytes();
				System.arraycopy(b, 0, buf, 0, b.length);
				fd.write(buf, 0, 128);
			}
			fd.close();
			return true;
		} catch (Exception ee) {
			ee.printStackTrace();
			return false;
		}
	}
}
