import java.io.*;

public class format {
	private static void help() {
		System.err.format("Usage: format [options] <file> [format-options]\n"+
			"       format show [options] [format-options]\n"+
			"       format list\n"+
			"Options:\n" +
			"    files=d  = preload image with files from directory 'd'\n" +
			"    h8d=file = preload image from H8D image\n" +
			"    imd=file = preload image from IMD image\n" +
			"    td0=file = preload image from TD0 image\n" +
			"    list     = List all known formats\n" +
			"    show     = Show geometry(s) (do not format)\n" +

			"Format-Options:\n" +
			"    5 | 8\n" +
			"    DD | SD\n" +
			"    DS | SS\n" +
			"    DT | ST\n" +
			"    MMS | Z17 | M47 | Z37 | Z47 | Z67 | Z37X | Z47X | Kaypro\n"
			);
		System.exit(1);
	}

	public static void main(String[] args) {
		File fd = null;
		// defaults:
		boolean list = false;
		boolean show = false;
		boolean ds = true;
		boolean dt = false;
		File preload = null;
		File h8d = null;
		File imd = null;
		File td0 = null;
		File raw = null;
		int fmt = H89FloppyFormatter.SD;
		boolean fmted = false;

		if (args.length < 1) {
			help(); // does not return
		}
		int x = 0;
		for (; x < args.length; ++x) {
			if (args[x].equalsIgnoreCase("list")) {
				list = true;
			} else if (args[x].equalsIgnoreCase("show")) {
				show = true;
			} else if (args[x].startsWith("imd=")) {
				File f = new File(args[x].substring(4));
				if (!f.exists() || f.isDirectory()) {
					System.err.format("No file %s\n", f.getAbsolutePath());
				} else {
					imd = f;
				}
			} else if (args[x].startsWith("td0=")) {
				File f = new File(args[x].substring(4));
				if (!f.exists() || f.isDirectory()) {
					System.err.format("No file %s\n", f.getAbsolutePath());
				} else {
					td0 = f;
				}
			} else if (args[x].startsWith("raw=")) {
				File f = new File(args[x].substring(4));
				if (!f.exists() || f.isDirectory()) {
					System.err.format("No file %s\n", f.getAbsolutePath());
				} else {
					raw = f;
				}
			} else if (args[x].startsWith("h8d=")) {
				File f = new File(args[x].substring(4));
				if (!f.exists() || f.isDirectory()) {
					System.err.format("No file %s\n", f.getAbsolutePath());
				} else {
					h8d = f;
				}
			} else if (args[x].startsWith("files=")) {
				File f = new File(args[x].substring(6));
				if (!f.exists() || !f.isDirectory()) {
					System.err.format("No directory %s\n", f.getAbsolutePath());
				} else {
					preload = f;
				}
			} else {
				break;
			}
		}
		if (list) {
			H89FloppyFormatter.list();
			System.exit(0);
		}
		String name = null;
		if (!show) {
			if (x >= args.length) {
				help(); // does not return
			}
			name = args[x++];
			if (!name.endsWith(".logdisk")) {
				name += ".logdisk";
			}

			fd = new File(name);
			if (fd.exists()) {
				System.err.format("File %s already exists\n", name);
				System.exit(1);
			}
		}
		while (x < args.length) {
			if (args[x].equalsIgnoreCase("5")) {
				fmt = (fmt & ~H89FloppyFormatter.MED_MASK) | H89FloppyFormatter.F525;
			} else if (args[x].equalsIgnoreCase("8")) {
				fmt = (fmt & ~H89FloppyFormatter.MED_MASK) | H89FloppyFormatter.F8;
			} else if (args[x].equalsIgnoreCase("DD")) {
				fmt = (fmt & ~H89FloppyFormatter.DEN_MASK) | H89FloppyFormatter.DD;
			} else if (args[x].equalsIgnoreCase("SD")) {
				fmt = (fmt & ~H89FloppyFormatter.DEN_MASK) | H89FloppyFormatter.SD;
			} else if (args[x].equalsIgnoreCase("DS")) {
				ds = true;
			} else if (args[x].equalsIgnoreCase("SS")) {
				ds = false;
			} else if (args[x].equalsIgnoreCase("DT")) {
				dt = true;
			} else if (args[x].equalsIgnoreCase("ST")) {
				dt = false;
			} else if (args[x].equalsIgnoreCase("HiDens")) {
				fmt = (fmt & ~H89FloppyFormatter.FMT_MASK) | H89FloppyFormatter.HIDENS;
			} else if (args[x].equalsIgnoreCase("Kaypro")) {
				fmt = (fmt & ~H89FloppyFormatter.FMT_MASK) | H89FloppyFormatter.KAYPRO;
			} else if (args[x].equalsIgnoreCase("MMS")) {
				fmt = (fmt & ~H89FloppyFormatter.FMT_MASK) | H89FloppyFormatter.MMS;
			} else if (args[x].equalsIgnoreCase("M47")) {
				fmt = (fmt & ~H89FloppyFormatter.FMT_MASK) | H89FloppyFormatter.M47;
				fmt = (fmt & ~H89FloppyFormatter.MED_MASK) | H89FloppyFormatter.F8;
			} else if (args[x].equalsIgnoreCase("Z17")) {
				fmt = (fmt & ~H89FloppyFormatter.FMT_MASK) | H89FloppyFormatter.Z17;
				fmt = (fmt & ~H89FloppyFormatter.MED_MASK) | H89FloppyFormatter.F525;
			} else if (args[x].equalsIgnoreCase("Z37")) {
				fmt = (fmt & ~H89FloppyFormatter.FMT_MASK) | H89FloppyFormatter.Z37;
				fmt = (fmt & ~H89FloppyFormatter.MED_MASK) | H89FloppyFormatter.F525;
			} else if (args[x].equalsIgnoreCase("Z47")) {
				fmt = (fmt & ~H89FloppyFormatter.FMT_MASK) | H89FloppyFormatter.Z47;
				fmt = (fmt & ~H89FloppyFormatter.MED_MASK) | H89FloppyFormatter.F8;
			} else if (args[x].equalsIgnoreCase("Z67")) {
				fmt = (fmt & ~H89FloppyFormatter.FMT_MASK) | H89FloppyFormatter.Z67;
				fmt = (fmt & ~H89FloppyFormatter.MED_MASK) | H89FloppyFormatter.F8;
			} else if (args[x].equalsIgnoreCase("Z37X")) {
				fmt = (fmt & ~H89FloppyFormatter.FMT_MASK) | H89FloppyFormatter.Z37X;
				fmt = (fmt & ~H89FloppyFormatter.MED_MASK) | H89FloppyFormatter.F525;
			} else if (args[x].equalsIgnoreCase("Z47X")) {
				fmt = (fmt & ~H89FloppyFormatter.FMT_MASK) | H89FloppyFormatter.Z47X;
				fmt = (fmt & ~H89FloppyFormatter.MED_MASK) | H89FloppyFormatter.F8;
			} else {
				System.err.format("Unsupported option: %s\n", args[x]);
				System.exit(1);
			}
			++x;
		}
		if ((fmt & H89FloppyFormatter.FMT_MASK) != 0 &&
				(fmt & H89FloppyFormatter.MED_MASK) == 0) {
			System.err.format("No media size specified\n");
			if (!show) {
				System.exit(1);
			}
		}
		fmted = !((fmt & H89FloppyFormatter.FMT_MASK) == 0 ||
				(fmt & H89FloppyFormatter.MED_MASK) == 0);
		if (!show && !fmted) {
			System.err.format("No format/media specified\n");
			System.exit(1);
		}

		if ((fmt & H89FloppyFormatter.MED_MASK) == H89FloppyFormatter.F8 && dt) {
			System.err.format("8\" DT not supported\n");
			System.exit(1);
		}
		H89FloppyFormatter ff = null;
		if (fmted) {
			ff = new H89FloppyFormatter(fmt, ds, dt);
			if (ff.isError()) {
				System.err.format("Unsupported format\n");
				System.exit(1);
			}
		}
		if (show) {
			// TODO: show geometries...
			if (h8d != null) {
				// try various geometries...
				// assume all are variants of Z17...
				long t = h8d.length() / 2560; // 10x256 sectors/track
				if (t == 40) {
					// SS ST
					System.err.format("%s: 5\" sides: 1 tracks: 40 sectors/track: 10 sector-size: 256\n", h8d.getName());                         
				} else if (t == 80) {
					// ST DS or DT SS
					System.err.format("%s: 5\" sides: 2 tracks: 40 sectors/track: 10 sector-size: 256\n", h8d.getName());                         
					System.err.format("%s: 5\" sides: 1 tracks: 80 sectors/track: 10 sector-size: 256\n", h8d.getName());                         
				} else if (t == 160) {
					// DT DS
					System.err.format("%s: 5\" sides: 2 tracks: 80 sectors/track: 10 sector-size: 256\n", h8d.getName());                         
				} else {
					System.err.format("%s: Invalid H8D image\n", h8d.getName());                         
				}
			}
			if (imd != null) {
				IMDDiskImage im = new IMDDiskImage(imd, false);
				if (im.isError()) {
					System.err.format("%s: Invalid IMD image\n", imd.getName());
				} else {
					System.err.format("%s: %d\" sides: %d tracks: %d sectors/track: %d sector-size: %d\n",
						imd.getName(), im.mediaSize(),
						im.numSides(), im.numTracks(),
						im.sectorsPerTrack(), im.sectorSize());
				}
			}
			if (td0 != null) {
				TD0DiskImage im = new TD0DiskImage(td0, false);
				if (im.isError()) {
					System.err.format("%s: Invalid TD0 image\n", td0.getName());
				} else {
					System.err.format("%s: %d\" sides: %d tracks: %d sectors/track: %d sector-size: %d\n",
						td0.getName(), im.mediaSize(),
						im.numSides(), im.numTracks(),
						im.sectorsPerTrack(), im.sectorSize());
				}
			}
			if (ff != null) {
				// print ff geometry...
				System.err.format("%s: %d\" sides: %d tracks: %d sectors/track: %d sector-size: %d\n",
					"Parameters", ff.mediaSize(),
					ff.numSides(), ff.numTracks(),
					ff.sectorsPerTrack(), ff.sectorSize());
			}
			System.exit(0);
		}

		if (!ff.format(fd)) {
			System.err.format("Format of %s failed\n", name);
			System.exit(1);
		}
		System.out.format("Formatted %s successfully.\n", name);
		if (h8d != null) {
			if (h8d.length() != ff.getCapacity()) {
				System.err.format("H8D image wrong size, "+
					"is %d expected %d\n",
					h8d.length(), ff.getCapacity());
				System.exit(1);
			}
			overlayImage(h8d, fd);
		} else if (raw != null) {
			if (raw.length() != ff.getCapacity()) {
				System.err.format("RAW image wrong size, "+
					"is %d expected %d\n",
					raw.length(), ff.getCapacity());
				System.exit(1);
			}
			overlayImage(raw, fd);
		} else if (imd != null) {
			IMDDiskImage im = new IMDDiskImage(imd, ff.interlaced());
			if (im.isError()) {
				System.err.format("Invalid IMD image\n");
				System.exit(1);
			}
			if (!ff.isCompatible(im)) {
				System.err.format("IMD image geometry incompatible\n");
				System.exit(1);
			}
			if (!im.overlay(fd)) {
				System.err.format("Failed to overlay IMD image\n");
				System.exit(1);
			}
			System.out.format("Overlayed IMD image from %s:\n",
				imd.getName());
			System.out.format("%s", im.getTitle().replaceAll("\r",""));
		} else if (td0 != null) {
			TD0DiskImage im = new TD0DiskImage(td0, ff.interlaced());
			if (im.isError()) {
				System.err.format("Invalid TD0 image\n");
				System.exit(1);
			}
			if (!ff.isCompatible(im)) {
				System.err.format("TD0 image geometry incompatible\n");
				System.exit(1);
			}
			if (!im.overlay(fd)) {
				System.err.format("Failed to overlay TD0 image\n");
				System.exit(1);
			}
			System.out.format("Overlayed TD0 image from %s:\n",
				td0.getName());
			System.out.format("%s\n", im.getTitle().replaceAll("\0","\n"));
		} else if (preload != null) {
			CpmCreateFile cf = new CpmCreateFile(fd);
			if (cf.isError()) {
				System.err.format("Skipping preloads due to CpmCreateFile error\n");
				System.exit(1);
			}
			cf.doPreload(preload, ff.dpb);
			cf.close();
		}
	}

	static private void overlayImage(File h8d, File fd) {
		byte[] img = new byte[(int)h8d.length()];
		try {
			FileInputStream fi = new FileInputStream(h8d);
			fi.read(img);
			fi.close();
		} catch (Exception ee) {
			System.err.format("Failed to read H8D image\n");
			return;
		}
		try {
			RandomAccessFile rf = new RandomAccessFile(fd, "rw");
			rf.seek(0);
			rf.write(img);
			rf.close();
		} catch (Exception ee) {
			System.err.format("Failed to overlay H8D on new image\n");
			return;
		}
	}
}
